__all__ = [
    "fit_background",
    "fit_integration",
    "fit_lowrank",
    "fit_polynomial",
    "fit_sparse",
]

# standard library
from collections.abc import Sequence

# dependencies
import numpy as np
import xarray as xr
from ndtools import Any, Range as NDRange
from scipy.ndimage import median_filter
from sklearn.decomposition import TruncatedSVD
from ..stats import mad, mean

# type hints
Range = tuple[float | None, float | None]

# constants
SIGMA_OVER_MAD = 1.4826


def fit_background(
    da: xr.DataArray,
    target: str,
    /,
    *,
    fit_per_array: bool = True,
    fit_per_observation: bool = True,
) -> xr.DataArray:
    """Fit background model to a DataArray.

    Args:
        da: DataArray to fit.
        target: Target state to fit.
        fit_per_array: Whether to fit per array.
        fit_per_observation: Whether to fit per observation.

    Returns:
        Modeled background DataArray.

    """
    groups: list[str] = []

    if fit_per_array:
        groups.append("array")

    if fit_per_observation:
        groups.append("observation")

    if groups:
        return da.groupby(groups).apply(
            fit_background,
            args=(target,),
            fit_per_array=False,
            fit_per_observation=False,
        )

    model = (
        da.sel(time=da.state == target)
        .groupby("scan")
        .apply(mean, dim="time", keep_dims=True)
    )

    if model.sizes["time"] > 1:
        return model.interp_like(da, kwargs={"fill_value": "extrapolate"})
    else:
        return model.reindex_like(da, method="nearest")


def fit_integration(
    da: xr.DataArray,
    sigma: xr.DataArray,
    /,
    fit_seed: int = 0,
) -> xr.DataArray:
    """Fit cumulative-integration model to a DataArray.

    Args:
        da: DataArray to fit.
        sigma: DataArray of noise level.
        fit_seed: Random seed for shuffling time samples.

    Returns:
        Modeled cumulative-integration DataArray.

    """
    # shuffle time samples
    np.random.seed(fit_seed)
    index = np.random.permutation(np.arange(da.sizes["time"]))
    da = da.isel(time=index)
    sigma = sigma.isel(time=index)

    # calculate cumulative information
    weight = sigma**-2
    signal = (da * weight).cumsum("time") / weight.cumsum("time")
    noise = weight.cumsum("time") ** -0.5
    exposure = weight.cumsum("time") ** 2 / (weight**2 / da.exposure).cumsum("time")

    return signal.assign_coords(
        exposure=exposure.assign_attrs(long_name="Effective exposure time"),
        noise=noise.assign_attrs(long_name="Expected noise level"),
    ).assign_attrs(
        long_name=r"$T_{\mathrm{A}}^{\ast}$",
        units=da.units,
    )


def fit_lowrank(
    da: xr.DataArray,
    /,
    *,
    fit_components: int = 5,
    fit_off_only: bool = False,
    fit_per_array: bool = True,
    fit_per_observation: bool = True,
) -> xr.DataArray:
    """Fit low-rank model to a DataArray.

    Args:
        da: DataArray to fit.
        fit_components: Number of low-rank components to fit.
        fit_off_only: Whether to use only OFF samples for fitting.
        fit_per_array: Whether to fit per array.
        fit_per_observation: Whether to fit per observation.

    Returns:
        Modeled low-rank DataArray.

    """
    groups: list[str] = []

    if fit_per_array:
        groups.append("array")

    if fit_per_observation:
        groups.append("observation")

    if groups:
        return da.groupby(groups).apply(
            fit_lowrank,
            fit_components=fit_components,
            fit_off_only=fit_off_only,
            fit_per_array=False,
            fit_per_observation=False,
        )

    model = TruncatedSVD(fit_components)
    model.fit(da.sel(time=da.state == "OFF") if fit_off_only else da)
    return xr.zeros_like(da) + model.inverse_transform(model.transform(da))


def fit_polynomial(
    da: xr.DataArray,
    /,
    *,
    fit_degree: int = 1,
    fit_per_array: bool = True,
    fit_per_observation: bool = True,
    fit_ranges: Sequence[Range] = ((None, None),),
) -> xr.DataArray:
    """Fit polynomial model to a DataArray.

    Args:
        da: DataArray to fit.
        fit_degree: Degree of the polynomial to fit.
        fit_per_array: Whether to fit per array.
        fit_per_observation: Whether to fit per observation.
        fit_ranges: Frequency ranges in GHz to use for fitting.

    Returns:
        Modeled polynomial DataArray.

    """
    groups: list[str] = []

    if fit_per_array:
        groups.append("array")

    if fit_per_observation:
        groups.append("observation")

    if groups:
        return da.groupby(groups).apply(
            fit_polynomial,
            fit_degree=fit_degree,
            fit_per_array=False,
            fit_per_observation=False,
            fit_ranges=fit_ranges,
        )

    fit_ranges = Any(NDRange(*args) for args in fit_ranges)
    da_fit = da.sel(chan=da.frequency == fit_ranges)
    model = da_fit.polyfit("chan", fit_degree)

    return (
        xr.polyval(da_fit.chan, model.polyfit_coefficients)
        .interp_like(da, kwargs={"fill_value": "extrapolate"})
        .assign_coords(fit_ranges=da.frequency == fit_ranges)
    )


def fit_sparse(
    da: xr.DataArray,
    /,
    *,
    fit_per_array: bool = False,
    fit_per_observation: bool = False,
    fit_prefilter: int = 3,
    fit_threshold: float = 3.0,
) -> xr.DataArray:
    """Fit sparse model to a DataArray.

    Args:
        da: DataArray to fit.
        fit_per_array: Whether to fit per array.
        fit_per_observation: Whether to fit per observation.
        fit_prefilter: Size of the median filter before signal detection.
        fit_threshold: Absolute S/N threshold for signal detection.

    Returns:
        Modeled sparse DataArray.

    """
    groups: list[str] = []

    if fit_per_array:
        groups.append("array")

    if fit_per_observation:
        groups.append("observation")

    if groups:
        return da.groupby(groups).apply(
            fit_sparse,
            fit_per_array=False,
            fit_per_observation=False,
            fit_prefilter=fit_prefilter,
            fit_threshold=fit_threshold,
        )

    signal = da.sel(time=da.state == "ON").mean("time")
    signal.data = median_filter(signal.data, fit_prefilter)
    noise = SIGMA_OVER_MAD * mad(signal)
    signal[abs(signal / noise) < fit_threshold] = 0
    return (da.state == "ON") * signal



__all__ = ["cli", "nro45m", "stats"]
__version__ = "0.4.0"

# dependencies
from logging import basicConfig
from fire import Fire
from . import nro45m, stats


def cli() -> None:
    """Run the command line interface."""
    basicConfig(
        datefmt="%Y-%m-%d %H:%M:%S",
        format="[%(asctime)s %(name)s %(funcName)s %(levelname)s] %(message)s",
    )

    Fire(
        {
            "nro45m": {
                "qlook": {
                    "otf": nro45m.qlook.otf,
                    "psw": nro45m.qlook.psw,
                },
            }
        }
    )



__all__ = ["Array", "read"]

# standard library
from os import PathLike
from typing import Literal, get_args

# dependencies
import sam45
import numpy as np
import pandas as pd
import xarray as xr
from ..stats import mean

# type hints
Array = Literal[
    # fmt: off
    "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8",
    "A9", "A10", "A11", "A12", "A13", "A14", "A15", "A16",
    "A17", "A18", "A19", "A20", "A21", "A22", "A23", "A24",
    "A25", "A26", "A27", "A28", "A29", "A30", "A31", "A32",
    # fmt: on
]

# constants
AMBIENT_TEMPERATURE = 273.0  # K
DATETIME_FORMAT = "%Y%m%d%H%M%S.%f"
TIMEZONE_JST = pd.Timedelta(9, "h")


def read(
    log: PathLike[str] | str,
    array: Array,
    /,
    *,
    chan_binning: int = 1,
    time_binning: int = 1,
) -> xr.DataArray:
    """Read a SAM45 log to create a DataArray of an array.

    Args:
        log: Path to the SAM45 log.
        array: Array name to read (A1-A32).
        chan_binning: Number of channels to bin together.
        time_binning: Number of time samples to bin together.

    Returns:
        DataArray of the array.

    """
    if array not in get_args(Array):
        raise ValueError(f"Invalid array name: {array!r}")

    # read OBS information
    obs = sam45.read.obs(log)[0]
    obs_index = int(array[1:]) - 1

    # read DAT information
    dat = sam45.read.dat(log)
    dat_index = dat["cary_name"] == array.encode()

    # calculate DataArray dimensions
    time = pd.Series(dat[dat_index]["cint_sttm"]).astype(np.str_)
    time = time.mask(~time.str.contains(r"\."), time + ".0000")
    time = pd.to_datetime(time, format=DATETIME_FORMAT) - TIMEZONE_JST  # UTC
    chan = np.arange(obs["ichannel"][obs_index])  # zero-based index

    # calculate DataArray coordinates
    x_0, x_1 = obs["dfqdat_ch"][obs_index] - 1  # zero-based index
    y_0, y_1 = obs["dfqdat_fq"][obs_index] * 1e-9  # GHz
    frequency = (y_0 - y_1) / (x_0 - x_1) * (chan - x_0) + y_0  # GHz
    state = pd.Series(dat[dat_index]["cscan_type"].astype(np.str_))
    state = state.replace(r"^([A-Z]+).*$", r"\1", regex=True)
    state = state.to_numpy().astype(np.str_)

    # calculate DataArray data
    att = np.where(state == "R", obs["iary_ifatt"][obs_index], 0)  # dB
    data = dat[dat_index]["fary_data"].astype(np.float64)
    data = (data - data[state == "ZERO"]) * (10 ** (att / 10))[:, np.newaxis]

    # create DataArray
    dataarray = xr.DataArray(
        data=data,
        dims=("time", "chan"),
        coords={
            "array": (
                "time",
                np.full(len(time), array).astype(np.str_),
                {"long_name": "Array name"},
            ),
            "beam": (
                "time",
                np.full(len(time), obs["imlt_no"][obs_index].astype(np.int64)),
                {"long_name": "Beam number"},
            ),
            "chan": (
                "chan",
                chan,
                {"long_name": "Channel number"},
            ),
            "exposure": (
                "time",
                np.full(len(time), obs["diptim"]).astype(np.float64),
                {"long_name": "Exposure time", "units": "s"},
            ),
            "frequency": (
                "chan",
                frequency.astype(np.float64),
                {"long_name": "Rest frequency", "units": "GHz"},
            ),
            "observation": (
                "time",
                np.full(len(time), obs["clog_id"].astype(np.str_)),
                {"long_name": "Observation ID"},
            ),
            "scan": (
                "time",
                dat[dat_index]["iline_no"].astype(np.int64),
                {"long_name": "Scan number"},
            ),
            "state": (
                "time",
                state,
                {"long_name": "Scan state"},
            ),
            "temperature": (
                "time",
                np.full(len(time), AMBIENT_TEMPERATURE).astype(np.float64),
                {"long_name": "Ambient temperature", "units": "K"},
            ),
            "time": (
                "time",
                time,
                {"long_name": "Observed time in UTC"},
            ),
            "width": (
                "chan",
                np.full(len(chan), obs["dbechwid"][obs_index]) * 1e-9,
                {"long_name": "Channel width", "units": "GHz"},
            ),
        },
        attrs={
            "object": str(obs["cobj_name"].astype(np.str_)),
            "observer": str(obs["cgroup"].astype(np.str_)),
            "project": str(obs["cproject"].astype(np.str_)),
        },
    )

    # perform time and channel binning
    dataarray = (
        dataarray.sel(time=dataarray.state != "ZERO")
        .groupby("state")
        .apply(
            mean,
            dim={
                "time": time_binning,
                "chan": chan_binning,
            },
        )
    )
    return (
        dataarray.assign_coords(
            beam=dataarray.beam.astype(np.int64),
            chan=dataarray.chan.astype(np.int64),
            exposure=dataarray.exposure * time_binning,
            scan=dataarray.scan.astype(np.int64),
            width=dataarray.width * chan_binning,
        )
        .sortby("time")
        .sortby("chan")
    )



__all__ = [
    "all",
    "any",
    "apply",
    "count",
    "first",
    "last",
    "mad",
    "max",
    "mean",
    "median",
    "min",
    "prod",
    "std",
    "sum",
    "var",
]

# standard library
from collections.abc import Callable, Hashable, Iterable, Sequence
from typing import Any, Literal

# dependencies
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.types import Dims

# type hints
Boundary = Literal["exact", "trim", "pad"]
Side = Literal["left", "right"]
Stat = Callable[..., Any] | str

# constants
FIRST_INDEX = 0
LAST_INDEX = -1
NUMERIC_KINDS = "biufcmM"


def apply(
    da: xr.DataArray,
    func: Stat,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) statistical operation to a DataArray.

    Args:
        da: Input DataArray.
        func: Function or name of the statistical operation (e.g. ``'mean'``).
        dim: Name(s) of the dimension(s) along which the statistical operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the statistical operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the statistical operation.
        keep_dims: Whether to keep dimensions in the statistical operation.
        **options: Other options to be passed to the statistical operation.

    Returns:
        DataArray that the (chunked) statistical operation is applied.

    """
    if isinstance(dim, dict):
        pass
    elif dim is ... or dim is None:
        dim = da.sizes
    elif isinstance(dim, str):
        dim = {dim: da.sizes[dim]}
    elif isinstance(dim, Iterable):
        dim = {d: da.sizes[d] for d in dim}

    coord_func: dict[Hashable, Stat] = {}

    if numeric_coord_func == "first":
        numeric_coord_func = _first
    elif numeric_coord_func == "last":
        numeric_coord_func = _last

    if nonnumeric_coord_func == "first":
        nonnumeric_coord_func = _first
    elif nonnumeric_coord_func == "last":
        nonnumeric_coord_func = _last

    for name, coord in da.coords.items():
        if coord.dtype.kind in NUMERIC_KINDS:
            coord_func[name] = numeric_coord_func
        else:
            coord_func[name] = nonnumeric_coord_func

    coarsened = da.coarsen(
        dim,
        boundary=boundary,
        coord_func=coord_func,
        side=side,
    )

    if isinstance(func, str):
        reduced = getattr(coarsened, func)(
            keep_attrs=keep_attrs,
            **options,
        )
    elif callable(func):
        reduced = coarsened.reduce(
            func=func,
            keep_attrs=keep_attrs,
            **options,
        )
    else:
        raise TypeError("Func must be either callable or string.")

    return reduced if keep_dims else reduced.squeeze()


def all(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``all`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``all``  operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``all``  operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``all`` operation.
        keep_dims: Whether to keep dimensions in the ``all`` operation.
        **options: Other options to be passed to the ``all`` operation.

    Returns:
        DataArray that the (chunked) ``all``  operation is applied.

    """
    return apply(
        da,
        "all",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        **options,
    )


def any(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``any`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``any`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``any`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``any`` operation.
        keep_dims: Whether to keep dimensions in the ``any`` operation.
        **options: Other options to be passed to the ``any`` operation.

    Returns:
        DataArray that the (chunked) ``any`` operation is applied.

    """
    return apply(
        da,
        "any",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        **options,
    )


def count(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``count`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``count`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``count`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``count`` operation.
        keep_dims: Whether to keep dimensions in the ``count`` operation.
        **options: Other options to be passed to the ``count`` operation.

    Returns:
        DataArray that the (chunked) ``count`` operation is applied.

    """
    return apply(
        da,
        "count",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        **options,
    )


def first(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``first`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``first`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``first`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``first`` operation.
        keep_dims: Whether to keep dimensions in the ``first`` operation.
        **options: Other options to be passed to the ``first`` operation.

    Returns:
        DataArray that the (chunked) ``first`` operation is applied.

    """
    return apply(
        da,
        _first,
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        **options,
    )


def last(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``last`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``last`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``last`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``last`` operation.
        keep_dims: Whether to keep dimensions in the ``last`` operation.
        **options: Other options to be passed to the ``last`` operation.

    Returns:
        DataArray that the (chunked) ``last`` operation is applied.

    """
    return apply(
        da,
        _last,
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        **options,
    )


def mad(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``mad`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``mad`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``mad`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``mad`` operation.
        keep_dims: Whether to keep dimensions in the ``mad`` operation.
        skipna: Whether to ignore missing values in the ``mad`` operation.
        **options: Other options to be passed to the ``mad`` operation.

    Returns:
        DataArray that the (chunked) ``mad`` operation is applied.

    """
    return apply(
        da,
        _mad,
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def max(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``max`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``max`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``max`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``max`` operation.
        keep_dims: Whether to keep dimensions in the ``max`` operation.
        skipna: Whether to ignore missing values in the ``max`` operation.
        **options: Other options to be passed to the ``max`` operation.

    Returns:
        DataArray that the (chunked) ``max`` operation is applied.

    """
    return apply(
        da,
        "max",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def mean(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``mean`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``mean`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``mean`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``mean`` operation.
        keep_dims: Whether to keep dimensions in the ``mean`` operation.
        skipna: Whether to ignore missing values in the ``mean`` operation.
        **options: Other options to be passed to the ``mean`` operation.

    Returns:
        DataArray that the (chunked) ``mean`` operation is applied.

    """
    return apply(
        da,
        "mean",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def median(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``median`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``median`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``median`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``median`` operation.
        keep_dims: Whether to keep dimensions in the ``median`` operation.
        skipna: Whether to ignore missing values in the ``median`` operation.
        **options: Other options to be passed to the ``median`` operation.

    Returns:
        DataArray that the (chunked) ``median`` operation is applied.

    """
    return apply(
        da,
        "median",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def min(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``min`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``min`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``min`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``min`` operation.
        keep_dims: Whether to keep dimensions in the ``min`` operation.
        skipna: Whether to ignore missing values in the ``min`` operation.
        **options: Other options to be passed to the ``min`` operation.

    Returns:
        DataArray that the (chunked) ``min`` operation is applied.

    """
    return apply(
        da,
        "min",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def prod(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``prod`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``prod`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``prod`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``prod`` operation.
        keep_dims: Whether to keep dimensions in the ``prod`` operation.
        skipna: Whether to ignore missing values in the ``prod`` operation.
        **options: Other options to be passed to the ``prod`` operation.

    Returns:
        DataArray that the (chunked) ``prod`` operation is applied.

    """
    return apply(
        da,
        "prod",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def std(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``std`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``std`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``std`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``std`` operation.
        keep_dims: Whether to keep dimensions in the ``std`` operation.
        skipna: Whether to ignore missing values in the ``std`` operation.
        **options: Other options to be passed to the ``std`` operation.

    Returns:
        DataArray that the (chunked) ``std`` operation is applied.

    """
    return apply(
        da,
        "std",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def sum(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``sum`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``sum`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``sum`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``sum`` operation.
        keep_dims: Whether to keep dimensions in the ``sum`` operation.
        skipna: Whether to ignore missing values in the ``sum`` operation.
        **options: Other options to be passed to the ``sum`` operation.

    Returns:
        DataArray that the (chunked) ``sum`` operation is applied.

    """
    return apply(
        da,
        "sum",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def var(
    da: xr.DataArray,
    /,
    *,
    dim: Dims | dict[Hashable, int] = None,
    boundary: Boundary = "trim",
    side: Side | dict[Hashable, Side] = "left",
    numeric_coord_func: Stat = "mean",
    nonnumeric_coord_func: Stat = "first",
    keep_attrs: bool | None = None,
    keep_dims: bool | None = None,
    skipna: bool | None = None,
    **options: Any,
) -> xr.DataArray:
    """Apply a (chunked) ``var`` operation to a DataArray.

    Args:
        da: Input DataArray.
        dim: Name(s) of the dimension(s) along which the ``var`` operation
            will be applied. If a dictionary such as ``{dim: size, ...}``
            is specified, then the ``var`` operation will be applied
            to every data chunk of given size.
        boundary: Same option as ``xarray.DataArray.coarsen`` but defaults to ``'trim'``.
        side: Same option as ``xarray.DataArray.coarsen`` and defualts to ``'left'``.
        numeric_coord_func: Function or name of the statistical operation
            for the numeric coordinates (bool, numbers, datetime, timedelta).
        nonnumeric_coord_func: Function or name of the statistical operation
            for the non-numeric coordinates (str, bytes, and general object).
        keep_attrs: Whether to keep attributes in the ``var`` operation.
        keep_dims: Whether to keep dimensions in the ``var`` operation.
        skipna: Whether to ignore missing values in the ``var`` operation.
        **options: Other options to be passed to the ``var`` operation.

    Returns:
        DataArray that the (chunked) ``var`` operation is applied.

    """
    return apply(
        da,
        "var",
        dim=dim,
        boundary=boundary,
        side=side,
        numeric_coord_func=numeric_coord_func,
        nonnumeric_coord_func=nonnumeric_coord_func,
        keep_attrs=keep_attrs,
        keep_dims=keep_dims,
        skipna=skipna,
        **options,
    )


def _first(
    array: NDArray[Any],
    axis: Sequence[int] | int | None = None,
    **kwargs: Any,
) -> NDArray[Any]:
    """Similar to numpy.take(array, 0, axis) but supports multiple axes."""
    if not isinstance(axis, Sequence):
        return np.take(array, FIRST_INDEX, axis=axis)

    slices: list[slice | int] = [slice(None)] * array.ndim

    for ax in axis:
        slices[ax] = FIRST_INDEX

    return array[tuple(slices)]


def _last(
    array: NDArray[Any],
    axis: Sequence[int] | int | None = None,
    **kwargs: Any,
) -> NDArray[Any]:
    """Similar to numpy.take(array, -1, axis) but supports multiple axes."""
    if not isinstance(axis, Sequence):
        return np.take(array, LAST_INDEX, axis=axis)

    slices: list[slice | int] = [slice(None)] * array.ndim

    for ax in axis:
        slices[ax] = LAST_INDEX

    return array[tuple(slices)]


def _mad(
    array: NDArray[Any],
    axis: Sequence[int] | int | None = None,
    **kwargs: Any,
) -> NDArray[Any]:
    """Internal implementation of median absolute deviation."""
    keepdims = kwargs.pop("keepdims", False)
    skipna = kwargs.pop("skipna", False)
    median = np.nanmedian if skipna else np.median

    return median(
        np.abs(array - median(array, axis=axis, keepdims=True)),
        axis=axis,
        keepdims=keepdims,
    )



__all__ = ["otf", "psw"]

# standard library
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from itertools import product
from logging import DEBUG, getLogger
from os import PathLike
from pathlib import Path
from warnings import catch_warnings, simplefilter

# dependencies
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from tqdm import tqdm
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.figure import Figure
from .models import (
    fit_background,
    fit_integration,
    fit_lowrank,
    fit_polynomial,
    fit_sparse,
)
from .io import Array, read
from .. import __version__ as demist_version

# type hints
Range = tuple[float | None, float | None]

# constants
LOGGER = getLogger(__name__)


@contextmanager
def set_logger(debug: bool, /) -> Iterator[None]:
    """Temporarily set the level of the module logger."""
    level = LOGGER.level

    if debug:
        LOGGER.setLevel(DEBUG)

    try:
        yield
    finally:
        LOGGER.setLevel(level)


def otf(
    *logs: PathLike[str] | str,
    # options for reading SAM45 logs
    arrays: Sequence[Array] = ("A1",),
    chan_binning: int = 8,
    time_binning: int = 5,
    # options for the PolyFit analysis
    polyfit_degree: int = 1,
    polyfit_ranges: Sequence[Range] = ((None, None),),
    # options for the DE:MIST analysis
    demist_iterations: int = 100,
    demist_threshold: float = 1e-4,
    lowrank_components: int = 5,
    lowrank_off_only: bool = False,
    lowrank_per_array: bool = True,
    lowrank_per_observation: bool = True,
    sparse_per_array: bool = False,
    sparse_per_observation: bool = False,
    sparse_prefilter: int = 3,
    sparse_threshold: float = 3.0,
    # options for saving the quick-look results
    figsize: tuple[float, float] = (10, 5),
    simple: bool = True,
    xlim: Range = (None, None),
    ylim: Range = (None, None),
    # options for displaying
    debug: bool = False,
    progress: bool = True,
) -> Path:
    """Quick-look at a DE:MIST on-the-fly (OTF) mapping observation.

    Args:
        *logs: Path(s) to SAM45 log(s).
        arrays: Array names to read (A1-A32).
        chan_binning: Number of channels to bin together.
        time_binning: Number of time samples to bin together.
        polyfit_degree: Degree of polynomial for the PolyFit analysis.
        polyfit_ranges: Frequency ranges in GHz to use for the PolyFit analysis.
        demist_iterations: Number of maximum iterations for the DE:MIST analysis.
        demist_threshold: Convergence threshold for the DE:MIST analysis.
        lowrank_components: Number of components for low-rank model fitting.
        lowrank_off_only: Whether to use only OFF samples for low-rank model fitting.
        lowrank_per_array: Whether to fit low-rank model per array.
        lowrank_per_observation: Whether to fit low-rank model per observation.
        sparse_per_array: Whether to fit sparse model per array.
        sparse_per_observation: Whether to fit sparse model per observation.
        sparse_prefilter: Size of median filter for sparse model fitting.
        sparse_threshold: Absolute S/N threshold for sparse model fitting.
        figsize: Size of the saved quick-look results.
        simple: Whether not to save miscellaneous information.
        xlim: X-axis limits for the saved quick-look results.
        ylim: Y-axis limits for the saved quick-look results.
        debug: Whether to display debug information.
        progress: Whether to display progress bar.

    Returns:
        Absolute path to the saved quick-look results.
        If multiple logs are given, the last log's name will be used for saving.

    """
    with set_logger(debug):
        for key, val in (params := locals().copy()).items():
            LOGGER.debug(f"{key}: {val!r}")

    raise NotImplementedError("This command is not yet implemented.")


def psw(
    *logs: PathLike[str] | str,
    # options for reading SAM45 logs
    arrays: Sequence[Array] = ("A1",),
    chan_binning: int = 8,
    time_binning: int = 5,
    # options for the PolyFit analysis
    polyfit_degree: int = 1,
    polyfit_ranges: Sequence[Range] = ((None, None),),
    # options for the DE:MIST analysis
    demist_iterations: int = 100,
    demist_threshold: float = 1e-4,
    lowrank_components: int = 5,
    lowrank_off_only: bool = False,
    lowrank_per_array: bool = True,
    lowrank_per_observation: bool = True,
    sparse_per_array: bool = False,
    sparse_per_observation: bool = False,
    sparse_prefilter: int = 3,
    sparse_threshold: float = 3.0,
    # options for saving the quick-look results
    figsize: tuple[float, float] = (10, 5),
    simple: bool = True,
    xlim: Range = (None, None),
    ylim: Range = (None, None),
    # options for displaying
    debug: bool = False,
    progress: bool = True,
) -> Path:
    """Quick-look at a DE:MIST position-switching (PSW) observation.

    Args:
        *logs: Path(s) to SAM45 log(s).
        arrays: Array names to read (A1-A32).
        chan_binning: Number of channels to bin together.
        time_binning: Number of time samples to bin together.
        polyfit_degree: Degree of polynomial for the PolyFit analysis.
        polyfit_ranges: Frequency ranges in GHz to use for the PolyFit analysis.
        demist_iterations: Number of maximum iterations for the DE:MIST analysis.
        demist_threshold: Convergence threshold for the DE:MIST analysis.
        lowrank_components: Number of components for low-rank model fitting.
        lowrank_off_only: Whether to use only OFF samples for low-rank model fitting.
        lowrank_per_array: Whether to fit low-rank model per array.
        lowrank_per_observation: Whether to fit low-rank model per observation.
        sparse_per_array: Whether to fit sparse model per array.
        sparse_per_observation: Whether to fit sparse model per observation.
        sparse_prefilter: Size of median filter for sparse model fitting.
        sparse_threshold: Absolute S/N threshold for sparse model fitting.
        figsize: Size of the saved quick-look results.
        simple: Whether not to save miscellaneous information.
        xlim: X-axis limits for the saved quick-look results.
        ylim: Y-axis limits for the saved quick-look results.
        debug: Whether to display debug information.
        progress: Whether to display progress bar.

    Returns:
        Absolute path to the saved quick-look results.
        If multiple logs are given, the last log's name will be used for saving.

    """
    with set_logger(debug):
        for key, val in (params := locals().copy()).items():
            LOGGER.debug(f"{key}: {val!r}")

    # Read SAM45 logs and arrays
    with tqdm(
        desc="Reading SAM45 logs/arrays",
        disable=not progress,
        total=len(logs) * len(arrays),
    ) as bar:
        Ps: list[xr.DataArray] = []

        for log, array in product(logs, arrays):
            P = read(
                log,
                array,
                time_binning=time_binning,
                chan_binning=chan_binning,
            )

            # swap ON/OFF for non-central beams
            is_on = (P.state == "ON") & (P.beam != 1)
            is_off = (P.state == "OFF") & (P.beam != 1)
            P.state[is_on] = "OFF"
            P.state[is_off] = "ON"

            Ps.append(P)
            bar.update(1)

        P = xr.concat(Ps, dim="time").sortby("time")

    # run PolyFit (conventional) analysis
    with tqdm(
        desc="Running PolyFit analysis",
        disable=not progress,
        total=1,
    ) as bar:
        P_cal = fit_background(P, "R").sel(time=P.state.isin(["ON", "OFF"]))
        P_sky = fit_background(P, "OFF").sel(time=P.state.isin(["ON", "OFF"]))
        P = P.sel(time=P.state.isin(["ON", "OFF"]))

        T: xr.DataArray = P.temperature * (P - P_sky) / (P_cal - P_sky)
        T_poly = fit_polynomial(
            T,
            fit_degree=polyfit_degree,
            fit_ranges=polyfit_ranges,
            fit_per_array=True,
            fit_per_observation=True,
        )
        bar.update(1)

    # run DE:MIST (proposed) analysis
    with tqdm(
        desc="Running DE:MIST analysis",
        disable=not progress,
        total=demist_iterations,
    ) as bar:
        X: xr.DataArray = np.log(-(P - P_cal) / P.temperature)
        X_sparse = xr.zeros_like(X)

        for n in range(demist_iterations):
            X_lowrank = fit_lowrank(
                X - X_sparse,
                fit_components=lowrank_components,
                fit_off_only=lowrank_off_only,
                fit_per_array=lowrank_per_array,
                fit_per_observation=lowrank_per_observation,
            )

            X_sparse_ = X_sparse
            X_sparse = fit_sparse(
                X - X_lowrank,
                fit_per_array=sparse_per_array,
                fit_per_observation=sparse_per_observation,
                fit_prefilter=sparse_prefilter,
                fit_threshold=sparse_threshold,
            )

            with catch_warnings():
                simplefilter("ignore", category=RuntimeWarning)
                eps = float(((X_sparse - X_sparse_) / X_sparse_).mean())

            if abs(eps) < demist_threshold:
                bar.update(demist_iterations - n)
                break
            else:
                bar.update(1)

    # calculate final spectra
    with tqdm(
        desc="Calculating final spectra",
        disable=not progress,
        total=1,
    ) as bar:
        T_polyfit: xr.DataArray = T - T_poly
        T_polyfit = T_polyfit.assign_attrs(long_name=r"$T_{\mathrm{A}}^{\ast}$")
        T_demist: xr.DataArray = P.temperature * (1 - np.exp(X - X_lowrank))
        T_demist = T_demist.assign_attrs(long_name=r"$T_{\mathrm{A}}^{\ast}$")

        T_sys: xr.DataArray = P.temperature / (P_cal / P_sky - 1)
        T_sys = T_sys.assign_attrs(long_name=r"$T_{\mathrm{sys}}$")
        T_sigma: xr.DataArray = T_sys * (T.exposure * T.width * 1e9) ** -0.5
        T_sigma = T_sigma.assign_attrs(long_name="Expected noise level")

        S_polyfit = fit_integration(
            T_polyfit.sel(time=T_polyfit.state == "ON"),
            T_sigma.sel(time=T_sigma.state == "ON") * np.sqrt(2),
        )
        S_demist = fit_integration(
            T_demist.sel(time=T_demist.state == "ON"),
            T_sigma.sel(time=T_sigma.state == "ON"),
        )
        bar.update(1)

    # plot and save results
    with (
        tqdm(
            desc="Plotting/saving results",
            disable=not progress,
            total=1 if simple else 2,
        ) as bar,
        PdfPages(name := Path(log).name + f".{'+'.join(arrays)}.qlook.psw.pdf") as pdf,
    ):
        keywords = [
            f"{key}={value}".replace(" ", "")
            for key, value in {
                "demist_version": demist_version,
                "demist_function": "demist.nro45m.qlook.psw",
                **params,
            }.items()
        ]

        pdfinfo = pdf.infodict()
        pdfinfo["Title"] = "DE:MIST Quick-Look Results (NRO 45m PSW)"
        pdfinfo["Keywords"] = ", ".join(keywords)

        pdf.savefig(
            plot_integrated_info(
                figsize=figsize,
                horizontal=True,
                S_demist=S_demist,
                S_polyfit=S_polyfit,
                xlim=xlim,
                ylim=ylim,
            )
        )
        pdf.savefig(
            plot_integrated_info(
                figsize=figsize,
                horizontal=False,
                S_demist=S_demist,
                S_polyfit=S_polyfit,
                xlim=xlim,
                ylim=ylim,
            )
        )
        bar.update(1)

        if simple:
            return Path(name).resolve()

        pdf.savefig(
            plot_cumulative_info(
                S_demist=S_demist,
                S_polyfit=S_polyfit,
                figsize=figsize,
            )
        )

        for fig in plot_timewise_info(
            figsize=figsize,
            T_demist=T_demist,
            T_polyfit=T_polyfit,
            T_sys=T_sys,
        ):
            pdf.savefig(fig)

        pdf.savefig(
            plot_chanwise_info(
                figsize=figsize,
                T_demist=T_demist,
                T_polyfit=T_polyfit,
            )
        )
        bar.update(1)
        return Path(name).resolve()


def cov(da: xr.DataArray, /) -> xr.DataArray:
    """Calculate the normalized covariance matrix."""
    da = da.swap_dims(chan="frequency")
    cov = np.cov(((da - da.mean("time")) / da.std("time")).T)
    cov[np.diag_indices_from(cov)] = np.nan

    return xr.DataArray(
        cov,
        dims=("frequency_0", "frequency_1"),
        coords={
            "frequency_0": da.frequency.rename(frequency="frequency_0"),
            "frequency_1": da.frequency.rename(frequency="frequency_1"),
        },
        attrs={
            "long_name": "Normalized covariance",
        },
    )


def edgena(da: xr.DataArray, /) -> xr.DataArray:
    """Set the edge time samples to N/A (for plotting)."""
    da.loc[{"time": da.time[[0, -1]]}] = np.nan
    return da


def plot_chanwise_info(
    *,
    T_demist: xr.DataArray,
    T_polyfit: xr.DataArray,
    figsize: tuple[float, float],
) -> Figure:
    """Plot channel-wise information (covariance matrices)."""
    fig, axes = plt.subplots(1, 2, figsize=figsize, sharey=True)

    ax = axes[0]
    cov_polyfit = cov(T_polyfit)
    cov_polyfit.plot.pcolormesh(
        ax=ax,
        cmap="coolwarm",
        rasterized=True,
        vmin=-3.0 * (sigma := float(cov_polyfit.std())),
        vmax=+3.0 * sigma,
    )
    ax.set_title("Normalized covariance (PolyFit)")

    ax = axes[1]
    cov_demist = cov(T_demist)
    cov_demist.plot.pcolormesh(
        ax=ax,
        cmap="coolwarm",
        rasterized=True,
        vmin=-3.0 * sigma,
        vmax=+3.0 * sigma,
    )
    ax.set_title("Normalized covariance (DE:MIST)")

    fig.tight_layout()
    return fig


def plot_cumulative_info(
    *,
    S_demist: xr.DataArray,
    S_polyfit: xr.DataArray,
    figsize: tuple[float, float],
) -> Figure:
    """Plot cumulative information (cumulative noise level and maximum S/N)."""
    fig, axes = plt.subplots(1, 2, figsize=figsize)
    fit_ranges = S_polyfit.fit_ranges

    ax = axes[0]
    ax.plot(
        S_polyfit.sel(chan=fit_ranges).exposure.mean("chan"),
        S_polyfit.sel(chan=fit_ranges).std("chan"),
        label="PolyFit",
    )
    ax.plot(
        S_demist.sel(chan=fit_ranges).exposure.mean("chan"),
        S_demist.sel(chan=fit_ranges).std("chan"),
        label="DE:MIST",
    )
    ax.set_title("Cumulative noise level")
    ax.set_xlabel("Effective exposure time [s]")
    ax.set_ylabel("Achieved noise level [K]")

    ax = axes[1]
    ax.plot(
        S_polyfit.sel(chan=fit_ranges).exposure.mean("chan"),
        S_polyfit.max("chan") / S_polyfit.sel(chan=fit_ranges).std("chan"),
        label="PolyFit",
    )
    ax.plot(
        S_demist.sel(chan=fit_ranges).exposure.mean("chan"),
        S_demist.max("chan") / S_demist.sel(chan=fit_ranges).std("chan"),
        label="DE:MIST",
    )
    ax.set_title("Maximum signal-to-noise ratio")
    ax.set_xlabel("Effective exposure time [s]")
    ax.set_ylabel("Maximum signal-to-noise ratio [1]")
    ax.set_ylim(1.0, None)

    for ax in axes:
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.margins(x=0.0)
        ax.legend()
        ax.grid()

    fig.tight_layout()
    return fig


def plot_integrated_info(
    *,
    figsize: tuple[float, float],
    horizontal: bool = True,
    S_demist: xr.DataArray,
    S_polyfit: xr.DataArray,
    xlim: Range = (None, None),
    ylim: Range = (None, None),
) -> Figure:
    """Plot integrated information (spectra with noise levels)."""
    if horizontal:
        fig, axes = plt.subplots(1, 2, figsize=figsize, sharex=True, sharey=True)
    else:
        fig, axes = plt.subplots(2, 1, figsize=figsize, sharex=True, sharey=True)

    alpha_polyfit = (
        # fmt: off
        S_polyfit.isel(time=-1).sel(chan=S_polyfit.fit_ranges).std()
        / S_demist.noise.isel(time=-1)
        # fmt: on
    )
    alpha_demist = (
        # fmt: off
        S_demist.isel(time=-1).sel(chan=S_polyfit.fit_ranges).std()
        / S_demist.noise.isel(time=-1)
        # fmt: on
    )
    spec_polyfit = S_polyfit.isel(time=-1).swap_dims(chan="frequency")
    spec_demist = S_demist.isel(time=-1).swap_dims(chan="frequency")

    ax = axes[0]
    spec_polyfit.plot.step(
        ax=ax,
        label=(
            # fmt: off
            "PolyFit "
            rf"($\alpha$ = {alpha_polyfit.mean():.2f} $\pm$ {alpha_polyfit.std():.2f})",
            # fmt: on
        ),
    )
    ax.fill_between(
        spec_polyfit.frequency,
        -spec_polyfit.noise,
        +spec_polyfit.noise,
        color="gray",
        alpha=0.25,
        ec="none",
        label=r"Expected noise level ($\alpha = \sqrt{2}$)",
    )
    ax.set_title("Integrated spectrum (PolyFit)")

    ax = axes[1]
    spec_demist.plot.step(
        ax=ax,
        label=(
            # fmt: off
            "DE:MIST "
            rf"($\alpha$ = {alpha_demist.mean():.2f} $\pm$ {alpha_demist.std():.2f})"
            # fmt: on
        ),
    )
    ax.fill_between(
        spec_demist.frequency,
        -spec_demist.noise,
        +spec_demist.noise,
        color="gray",
        alpha=0.25,
        ec="none",
        label=r"Expected noise level ($\alpha = 1$)",
    )
    ax.set_title("Integrated spectrum (DE:MIST)")

    for ax in axes:
        ax.margins(x=0.0)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        ax.legend()
        ax.grid()

    fig.tight_layout()
    return fig


def plot_timewise_info(
    *,
    figsize: tuple[float, float],
    T_demist: xr.DataArray,
    T_polyfit: xr.DataArray,
    T_sys: xr.DataArray,
) -> list[Figure]:
    """Plot time-wise information (time-frequency plots)."""
    figs: list[Figure] = []

    for group in T_polyfit.groupby("array"):
        fig, ax = plt.subplots(figsize=figsize)

        array, T_polyfit_ = group
        T_polyfit_ = T_polyfit_.groupby("observation").apply(edgena)
        T_polyfit_.swap_dims(chan="frequency").T.plot.pcolormesh(
            ax=ax,
            cmap="coolwarm",
            rasterized=True,
            vmin=-5.0 * (sigma := float(T_polyfit.std())),
            vmax=+5.0 * sigma,
        )
        ax.set_title(r"$T_{\mathrm{A}}^{\ast}$ " f"(PolyFit, {array})")
        fig.tight_layout()
        figs.append(fig)

    for group in T_demist.groupby("array"):
        fig, ax = plt.subplots(figsize=figsize)

        array, T_demist_ = group
        T_demist_ = T_demist_.groupby("observation").apply(edgena)
        T_demist_.swap_dims(chan="frequency").T.plot.pcolormesh(
            ax=ax,
            cmap="coolwarm",
            rasterized=True,
            vmin=-5.0 * sigma,
            vmax=+5.0 * sigma,
        )
        ax.set_title(r"$T_{\mathrm{A}}^{\ast}$ " f"(DE:MIST, {array})")
        fig.tight_layout()
        figs.append(fig)

    for group in T_sys.groupby("array"):
        fig, ax = plt.subplots(figsize=figsize)

        array, T_sys_ = group
        T_sys_ = T_sys_.groupby("observation").apply(edgena)
        T_sys_.swap_dims(chan="frequency").T.plot.pcolormesh(
            ax=ax,
            cmap="inferno",
            rasterized=True,
            vmin=float(T_sys.min()),
            vmax=float(T_sys.max()),
        )
        ax.set_title(r"$T_{\mathrm{sys}}$ " f"({array})")
        fig.tight_layout()
        figs.append(fig)

    return figs
